import jax
import jax.numpy as np


def fft(x):
    return np.fft.fft(x)


def ifft(x):
    return np.fft.ifft(x)


def fft_2d(x):
    return np.fft.fft2(x)


def ifft_2d(x):
    return np.fft.ifft2(x)


def approx_inverse(x):
    x = np.flip(x, axis=-1)
    return np.roll(x, 1, axis=-1)


def inverse_2d(x):
    x = ifft_2d(1. / fft_2d(x)).real
    return np.nan_to_num(x)


def projection(x):
    f = np.abs(fft(x))
    p = np.real(ifft(fft(x) / f))
    return np.nan_to_num(p)


def projection_2d(x):
    f = np.abs(fft_2d(x))
    p = ifft_2d(fft_2d(x) / f).real
    return np.nan_to_num(p)


def binding(x, y):
    b = ifft(fft(x) * fft(y))
    return np.real(b)


def binding_2d(x, y):
    b = ifft_2d(np.multiply(fft_2d(x), fft_2d(y)))
    return np.real(b)


def unbinding(s, y):
    yt = approx_inverse(y)
    return binding(s, yt)


def unbinding_2d(b, y):
    yt = inverse_2d(y)
    return binding_2d(b, yt)


def normal(shape, seed=0):
    d = np.prod(np.asarray(shape[1:]))
    std = 1. / np.sqrt(d)
    return std * jax.random.normal(jax.random.PRNGKey(seed), shape, dtype=np.float32)


def inner_product(x, y, axis=-1, keepdims=False):
    return np.sum(x * y, axis=axis, keepdims=keepdims, )


def cosine_similarity(x, y, axis=None, keepdims=None):
    if not axis:
        axis = tuple(range(-len(x.size()) // 2, 0))
    norm_x = np.linalg.norm(x, axis=axis, keepdims=keepdims)
    norm_y = np.linalg.norm(y, axis=axis, keepdims=keepdims)
    return np.sum(x * y, axis=axis, keepdims=keepdims) / (norm_x * norm_y)



import jax
import jax.numpy as np
import flax.linen as nn


class FFT(nn.Module):
    @nn.compact
    def __call__(self, x):
        return np.fft.fft(x)


class IFFT(nn.Module):
    @nn.compact
    def __call__(self, x):
        return np.fft.ifft(x)


class ApproxInverse(nn.Module):
    @nn.compact
    def __call__(self, x):
        x = np.flip(x, axis=-1)
        return np.roll(x, 1, axis=-1)


class Projection(nn.Module):
    def setup(self):
        self.fft = FFT()
        self.ifft = IFFT()

    @nn.compact
    def __call__(self, x):
        f = np.abs(self.fft(x))
        p = np.real(self.ifft(self.fft(x) / f))
        return np.nan_to_num(p)


class Binding(nn.Module):
    def setup(self):
        self.fft = FFT()
        self.ifft = IFFT()

    @nn.compact
    def __call__(self, x, y):
        b = self.ifft(self.fft(x) * self.fft(y))
        return np.real(b)


class Unbinding(nn.Module):
    def setup(self):
        self.approx_inverse = ApproxInverse()
        self.binding = Binding()

    @nn.compact
    def __call__(self, b, y):
        yt = self.approx_inverse(y)
        return self.binding(b, yt)


class CosineSimilarity(nn.Module):
    @nn.compact
    def __call__(self, x, y, axis=-1, keepdims=False):
        norm_x = np.linalg.norm(x, axis=axis, keepdims=keepdims)
        norm_y = np.linalg.norm(y, axis=axis, keepdims=keepdims)
        return np.sum(x * y, axis=axis, keepdims=keepdims) / (norm_x * norm_y)


def normal(shape, seed=0):
    d = np.prod(np.asarray(shape[1:]))
    std = 1. / np.sqrt(d)
    return std * jax.random.normal(jax.random.PRNGKey(seed), shape, dtype=np.float32)



import tensorflow as tf


def fft(x):
    return tf.signal.rfft(x)


def ifft(x):
    return tf.signal.irfft(x)


def fft_2d(x):
    return tf.signal.rfft2d(x)


def ifft_2d(x):
    return tf.signal.irfft2d(x)


def approx_inverse(x):
    x = tf.reverse(x, axis=[-1])
    return tf.roll(x, 1, axis=-1)


def inverse_2d(x):
    x = ifft_2d(1. / fft_2d(x))
    return tf.where(tf.math.is_nan(x), 0., x)


def projection(x):
    fx = fft(x)
    p = ifft(fx / tf.cast(tf.abs(fx), dtype=tf.complex64))
    return tf.where(tf.math.is_nan(p), 0., p)


def projection_2d(x):
    fx = fft_2d(x)
    p = ifft_2d(fx / tf.cast(tf.abs(fx), dtype=tf.complex64))
    return tf.where(tf.math.is_nan(p), 0., p)


def binding(x, y):
    return ifft(fft(x) * fft(y))


def binding_2d(x, y):
    return ifft_2d(tf.multiply(fft_2d(x), fft_2d(y)))


def unbinding(s, y):
    yt = approx_inverse(y)
    return binding(s, yt)


def unbinding_2d(b, y):
    yt = inverse_2d(y)
    return binding_2d(b, yt)


def normal(shape, seed):
    tf.random.set_seed(seed)
    d = tf.reduce_prod(tf.convert_to_tensor(shape[1:], dtype=tf.float32))
    return tf.random.normal(shape=shape, mean=0., stddev=1. / tf.sqrt(d))


def inner_product(x, y, axis=-1, keepdims=False):
    return tf.reduce_sum(x * y, axis=axis, keepdims=keepdims)


def cosine_similarity(x, y, axis=None, keepdims=None):
    if not axis:
        axis = tuple(range(-len(x.size()) // 2, 0))
    norm_x = tf.norm(x, axis=axis, keepdims=keepdims)
    norm_y = tf.norm(y, axis=axis, keepdims=keepdims)
    return tf.reduce_sum(x * y, axis=axis, keepdims=keepdims) / (norm_x * norm_y)


if __name__ == '__main__':
    x = normal(shape=(2, 4), seed=0)
    print(x)
    print(projection_2d(x))



import torch


def fft(x):
    return torch.fft.fft(x)


def ifft(x):
    return torch.fft.ifft(x)


def fft_2d(x):
    return torch.fft.fft2(x)


def ifft_2d(x):
    return torch.fft.ifft2(x)


def approx_inverse(x):
    x = torch.flip(x, dims=[-1])
    return torch.roll(x, 1, dims=-1)


def inverse_2d(x):
    x = ifft_2d(1. / fft_2d(x)).real
    return torch.nan_to_num(x)


def projection(x):
    f = torch.abs(fft(x))
    p = ifft(fft(x) / f).real
    return torch.nan_to_num(p)


def projection_2d(x):
    f = torch.abs(fft_2d(x))
    p = ifft_2d(fft_2d(x) / f).real
    return torch.nan_to_num(p)


def binding(x, y):
    s = ifft(torch.multiply(fft(x), fft(y)))
    return s.real


def binding_2d(x, y):
    b = ifft_2d(torch.multiply(fft_2d(x), fft_2d(y)))
    return b.real


def unbinding(b, y):
    yt = approx_inverse(y)
    return binding(b, yt)


def unbinding_2d(b, y):
    yt = inverse_2d(y)
    return binding_2d(b, yt)


def normal(shape, seed):
    torch.manual_seed(seed)
    d = torch.prod(torch.tensor(shape[1:]))
    return torch.normal(mean=0, std=1. / torch.sqrt(d), size=shape)


def inner_product(x, y, dim=-1, keepdim=False):
    return torch.sum(x * y, dim=dim, keepdim=keepdim)


def cosine_similarity(x, y, dim=None, keepdim=False):
    if not dim:
        dim = list(range(-len(x.size()) // 2, 0))
    norm_x = torch.norm(x, dim=dim, keepdim=keepdim)
    norm_y = torch.norm(y, dim=dim, keepdim=keepdim)
    return torch.sum(x * y, dim=dim, keepdim=keepdim) / (norm_x * norm_y)



import re
import setuptools

with open('HRR/__init__.py', 'r', encoding='utf-8') as fd:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE
    ).group(1)

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

setuptools.setup(name='hrr',
                 version=version,
                 author='Mohammad Mahmudul Alam',
                 description='Holographic Reduced Representations',
                 long_description=long_description,
                 long_description_content_type='text/markdown',
                 classifiers=[
                     'Programming Language :: Python :: 3',
                     'Operating System :: OS Independent',
                 ],
                 packages=setuptools.find_packages(),
                 install_requires=[],
                 python_requires='>=3.7',
                 )
